// C Source File
// Created 7/7/2011; 12:28:31 PM

// Memory Usage for the TI-89 by TC01
// Ultimate goal is to provide a nice GUI for showing what's taking up space on the system

#define COMMENT_STRING         "Gathers information about memory"
#define COMMENT_PROGRAM_NAME   "Memory Usage"
#define COMMENT_VERSION_NUMBER 0,0,3,0
#define COMMENT_AUTHORS        "TC01"

#include <alloc.h>
#include <asmtypes.h>
#include <args.h>
#include <dialogs.h>
#include <error.h>
#include <flash.h>
#include <flashapp.h>
#include <kbd.h>
#include <mem.h>
#include <menus.h>
#include <stdio.h>
#include <stdlib.h>
#include <statline.h>
#include <string.h>
#include <timath.h>
#include <vat.h>
#include <unknown.h>

#include "apps.h"

//Symbol structure for linked list (this is a little messy, some tags are folder only, some are file only)
typedef struct LLSymbol
{
	struct LLSymbol* next;
	struct LLSymbol* previous;
	char name[10];

	//Folders only
	int isFolder;
	int numFiles;
	
	//Files only
	int isArchived;
} LLSymbol;

//Memory info structure for a folder
typedef struct FolderMemInfo
{
	char name[10];
	int numFiles;
	unsigned long size;
	unsigned long archivedSize;
	unsigned long unarchivedSize;
} FolderMemInfo;

//Memory info structure for a file
typedef struct FileMemInfo
{
	char name[10];
	unsigned long size;
	int isArchived;
} FileMemInfo;

//A function to get general memory information and print it
void doGeneralInformation()
{
	int result;
	unsigned long inUse;
	unsigned long freeMem;
	unsigned long freedByGC;
	unsigned long unusedSectors;
	unsigned long badSectors;
	unsigned long allExceptBaseCode;
	unsigned long freeRAM;
	char* buffer = NULL;
	HANDLE dlg = H_NULL;
	HANDLE dlg2 = H_NULL;
	
	//Call the information functions and allocate memory
	freeRAM = HeapAvail();
	EM_survey(&inUse, &freedByGC, &freeMem, &unusedSectors, &badSectors, &allExceptBaseCode);
	buffer = (char*)calloc(40, sizeof(char));
	
	//Create a dialog with RAM information
	dlg = DialogNewSimple(140, 45);
	DialogAddTitle(dlg, "RAM Info", BT_OK, BT_CANCEL);

	sprintf(buffer, "Used: %lu bytes", 192512 - freeRAM);
	DialogAddText(dlg, 5, 15, buffer);
	memset(buffer, '\0', strlen(buffer));	
	
	sprintf(buffer, "Free: %lu bytes", freeRAM);
	DialogAddText(dlg, 5, 25, buffer);
	memset(buffer, '\0', strlen(buffer));
	
	//Create a dialog with the flash archive information
	dlg2 = DialogNewSimple(140, 55);
	DialogAddTitle(dlg2, "Flash Archive Info", BT_OK, BT_CANCEL);
	
	sprintf(buffer, "Used: %lu bytes", inUse);
	DialogAddText(dlg2, 5, 15, buffer);
	memset(buffer, '\0', strlen(buffer));
	
	sprintf(buffer, "Free: %lu bytes", freeMem);
	DialogAddText(dlg2, 5, 25, buffer);
	memset(buffer, '\0', strlen(buffer));
	
	sprintf(buffer, "Garbage: %lu bytes", freedByGC);
	DialogAddText(dlg2, 5, 35, buffer);

	//Display the dialogs and then free memory
	result = DialogDo(dlg, CENTER, CENTER, NULL, NULL);
	result = DialogDo(dlg2, CENTER, CENTER, NULL, NULL);
	HeapFree(dlg);
	HeapFree(dlg2);
	free(buffer);
}

//Compares folder mem infos by comparing their sizes (greatest -> smallest)
int folderCompare(const void* a, const void* b)
{
	const FolderMemInfo* aFolder = a;
	const FolderMemInfo* bFolder = b;
	int comparison = fcmp(aFolder->size, bFolder->size);
	return -comparison;
}

//Compares file mem infos by comparing their sizes (greatest -> smallest)
int fileCompare(const void* a, const void* b)
{
	const FileMemInfo* aFolder = a;
	const FileMemInfo* bFolder = b;
	int comparison = fcmp(aFolder->size, bFolder->size);
	return -comparison;
}

//A function to return the number of files in a folder
int getNumFilesInFolder(char* folderName)
{
	int numFiles = 0;
	char moddedFolderName[10];
	SYM_ENTRY* symData;

	//Lock the VAT folder table, mod folder name
	FolderOp(SYMSTR("\x7F"), FOP_LOCK);
	strncat(moddedFolderName, folderName, strlen(folderName) - 1);
	
	//Loop through the VAT and count
	symData = SymFindFirst(SYMSTR(moddedFolderName), FO_RECURSE);
	while (symData)
	{
		numFiles += 1;
		symData = SymFindNext();
	}
	
	//Unlock the VAT folder table
	FolderOp(SYMSTR("\x7F"), FOP_UNLOCK);
	
	//Return the number of files
	return numFiles;
}

//A function to load an array of FileMemInfos with information from files in the folder
void getFileInfos(FileMemInfo* fileMems, int numFiles, char* folderName)
{
	int i;
	int counter = 0;
	int fileSize;
	char* buffer = NULL;
	char moddedFolderName[10];
	SYM_ENTRY* symData;
	LLSymbol* fileSymbols = NULL;

	//Lock the VAT folder table, mod folder name, calloc memory
	FolderOp(SYMSTR("\x7F"), FOP_LOCK);
	memset(moddedFolderName, '\0', 10);
	strncpy(moddedFolderName, folderName, strlen(folderName) - 1);
	fileSymbols = (LLSymbol*)calloc(numFiles, sizeof(LLSymbol));
	buffer = (char*)calloc(17, sizeof(char));
	
	//Loop through the VAT and count
	symData = SymFindFirst(SYMSTR(moddedFolderName), FO_SINGLE_FOLDER);
	while (symData)
	{
		if (symData->flags.bits.folder != 1)
		{
			strncat(fileSymbols[counter].name, symData->name, 8);
			if (symData->flags.bits.archived == 1)
				fileSymbols[counter].isArchived = 1;
		}
		symData = SymFindNext();
		counter += 1;
	}
	
	//Loop through the LLSymbol infos
	for (i = 0; i < counter; i++)
	{
		buffer = strcpy(buffer, folderName);
		buffer = strcat(buffer, fileSymbols[i].name);
		strcpy(fileMems[i].name, fileSymbols[i].name);
		symData = SymFindPtr(SYMSTR(buffer), 0);
		fileSize = ((MULTI_EXPR*)HeapDeref(symData->handle))->Size + 2;
		fileMems[i].size += fileSize;
		if (fileSymbols[i].isArchived == 1)
			fileMems[i].isArchived = 1;
		if (symData->flags.bits.twin == 1)
			fileMems[i].size = 0;
		buffer = memset(buffer, '\0', strlen(buffer));
	}
	
	//Unlock the VAT folder table, free memory
	FolderOp(SYMSTR("\x7F"), FOP_UNLOCK);
	free(fileSymbols);
	free(buffer);
}

//A function to output the folder/file memory in nice menus
void displayFolderMemory(FolderMemInfo* folderMems, int numFolders)
{
	int i;
	int finished = 0;
	int numFiles = 0;
	short selected;
	short selectedSecond;
	short selectedThird;
	char* mString = NULL;
	
	HANDLE menu = H_NULL;
	HANDLE menu2 = H_NULL;
	HANDLE menu3 = H_NULL;
	FileMemInfo* fileMems = NULL;

	//Allocate memory
	mString = (char*)calloc(50, sizeof(char));
	
	//Create the menu
	menu = PopupNew("Folders", 0);
	qsort(folderMems, numFolders, sizeof(FolderMemInfo), folderCompare); 
	for (i = 0; i < numFolders; i++)
	{
		sprintf(mString, "%s: %lu", folderMems[i].name, folderMems[i].size);
		PopupAddText(menu, -1, mString, 0);
		memset(mString, '\0', strlen(mString));
	}

	//Display the menu, clean up the menu, and free memory
	selected = PopupDo(menu, CENTER, CENTER, 0);
	HeapFree(menu);
	if (selected)
	{
		selected -= 1;
		while (!finished)
		{
			menu2 = PopupNew(folderMems[selected].name, 0);
			sprintf(mString, "All: %lu", folderMems[selected].size);
			PopupAddText(menu2, -1, mString, 0);
			memset(mString, '\0', strlen(mString));
			
			sprintf(mString, "Archived: %lu", folderMems[selected].archivedSize);
			PopupAddText(menu2, -1, mString, 0);
			memset(mString, '\0', strlen(mString));
			
			sprintf(mString, "Unarchived: %lu", folderMems[selected].unarchivedSize);
			PopupAddText(menu2, -1, mString, 0);
			
			selectedSecond = PopupDo(menu2, CENTER, CENTER, 0);
			HeapFree(menu2);
			if (!selectedSecond)
			{
				displayFolderMemory(folderMems, numFolders);
				finished = 1;				
			}
			else
			{
				//Get file info (from some functions) and then sort and make a list of them
				finished = 0;
				mString = strcpy(mString, folderMems[selected].name);
				if (selectedSecond == 2)
					mString = strcat(mString, " Archived");
				if (selectedSecond == 3)
					mString = strcat(mString, " Unarchived");
				menu3 = PopupNew(mString, 0);
				memset(mString, '\0', strlen(mString));
				numFiles = getNumFilesInFolder(folderMems[selected].name);
				fileMems = (FileMemInfo*)calloc(numFiles, sizeof(FileMemInfo));
				getFileInfos(fileMems, numFiles, folderMems[selected].name);
				qsort(fileMems, numFiles, sizeof(FileMemInfo), fileCompare);
				for (i = 0; i < numFiles; i++)
				{
					if (fileMems[i].size == 0)
						continue;
					if (selectedSecond == 2)
					{
						if (fileMems[i].isArchived != 1)
							continue;
					}
					else if (selectedSecond == 3)
					{
						if (fileMems[i].isArchived == 1)
							continue;
					}
					sprintf(mString, "%s: %lu", fileMems[i].name, fileMems[i].size);
					PopupAddText(menu3, -1, mString, 0);
					memset(mString, '\0', strlen(mString));
				}

				//Display the menu, free memory
				selectedThird = PopupDo(menu3, CENTER, CENTER, 0);
				HeapFree(menu3);
				free(fileMems);
				if (selectedThird)
					finished = 1;
			}			
		}
	}
	free(mString);
}

//A function to get memory for each file and folder and output it
void doFilesAndFolders()
{
	int numFolders = 0;
	int counter = -1;
	int fileSize;
	char* currentFolder;
	char* buffer;

	LLSymbol* anchor;
	LLSymbol* temp;
	LLSymbol* folderPointer = NULL;
	FolderMemInfo* folderMems;
	SYM_ENTRY* symData;
	
	//Lock the VAT folder table
	FolderOp(SYMSTR("\x7F"), FOP_LOCK);
	
	//Run through the VAT to load all the names into the arrays
	symData = SymFindFirst("", FO_RECURSE);
	anchor = (LLSymbol*)calloc(sizeof(LLSymbol), 1);
	temp = anchor;
	temp->next = NULL;
	temp->previous = NULL;
	while (symData)
	{
		strncat(temp->name, symData->name, 8);
		if (symData->flags.bits.folder == 1)
		{
			temp->isFolder = 1;
			numFolders++;
			folderPointer = temp;
		}
		else
		{
			if (symData->flags.bits.twin == 1)
				continue;
			folderPointer->numFiles += 1;
			if (symData->flags.bits.archived == 1)
				temp->isArchived = 1;
		}
		symData = SymFindNext();
		if ((temp->next = (LLSymbol*)calloc(sizeof(LLSymbol), 1)) == NULL)
		{
			ST_helpMsg("Allocation error!");
			return;
		}
		temp->next->previous = temp;
		temp = temp->next;
		temp->next = NULL;
	}
		
	//Unlock the VAT folder table. If we don't get here, the folder table will stay locked
	FolderOp(SYMSTR("\x7F"), FOP_UNLOCK);
	
	//Alloc some arrays
	folderMems = (FolderMemInfo*)calloc(numFolders, sizeof(FolderMemInfo));
	buffer = (char*)calloc(17, sizeof(char));
	currentFolder = (char*)calloc(9, sizeof(char));
	
	//Main loop through
	temp = anchor;
	while (temp != NULL)
	{
		//If we've somehow opened something with no name, ignore it
		if (temp->name == NULL)
		{
			temp = temp->next;
			continue;
		}
		//If it's a folder, increment the counter and current folder var
		if (temp->isFolder == 1)
		{
			counter += 1;
			currentFolder = strcpy(currentFolder, temp->name);
			currentFolder = strcat(currentFolder, "\\");
			strcpy(folderMems[counter].name, currentFolder);
			folderMems[counter].numFiles = temp->numFiles;
			folderMems[counter].size = 0;
			ST_helpMsg(currentFolder);
		}
		//If it's not, get the size, and add them to the system
		else
		{
			buffer = strcpy(buffer, currentFolder);
			buffer = strcat(buffer, temp->name);
			symData = SymFindPtr(SYMSTR(buffer), 0);
			fileSize = ((MULTI_EXPR*)HeapDeref(symData->handle))->Size + 2;
			folderMems[counter].size += fileSize;
			if (temp->isArchived == 1)
				folderMems[counter].archivedSize += fileSize;
			else
				folderMems[counter].unarchivedSize += fileSize;
			buffer = memset(buffer, '\0', strlen(buffer));
		}
		temp = temp->next;
	}

	//Free the linked list!
	temp = anchor;
	while (temp->next != NULL)
		temp = temp->next;
	while (temp->previous != NULL)
	{
		temp = temp->previous;
		free(temp->next);
	}
	free(temp);

	//Display the memory info
	displayFolderMemory(folderMems, numFolders);
	
	//Free other memory
	free(folderMems);
	free(buffer);
	free(currentFolder);
}

//Compares app IDs by their size (greatest -> smallest)
int appIDCompare(const void* a, const void* b)
{
	const short* aAppID = (short*)a;
	const short* bAppID = (short*)b;
	int comparison = fcmp(OO_GetFlashAppSize(*aAppID), OO_GetFlashAppSize(*bAppID));
	return -comparison;
}

//A function to get the memory for each flash app and output it
void doFlashApps()
{
	int i;
	int numApps = 0;
	short appID;
	short selected;
	short dlgKey;
	short* appIDs = NULL;
	char* mString = NULL;
	char working[10];
	
	HANDLE menu;
	
	//Do one run through of the ACBs to get the number of non-builtin apps
	for (appID = OO_firstACB; appID != H_NULL; appID = OO_NextACB(appID))
	{
		if (OO_GetFlashAppSize(appID) > 4)
			numApps += 1;
	}
	
	//Alloc memory
	mString = (char*)calloc(100, sizeof(char));
	appIDs = (short*)calloc(numApps, sizeof(short));
	numApps = 0;
	
	//Do another run through and append app IDs to the array
	for (appID = OO_firstACB; appID != H_NULL; appID = OO_NextACB(appID))
	{
		if (OO_GetFlashAppSize(appID) > 4)
		{
			appIDs[numApps] = appID;
			numApps += 1;
		}
	}
	
	//Now, sort and build a menu
	menu = PopupNew("Applications", 0);
	qsort(appIDs, numApps, sizeof(short), appIDCompare);
	for (i = 0; i < numApps; i++)
	{
		if (strlen(GetAppName(appIDs[i])) > 12)
		{
			mString = strncpy(mString, GetAppName(appIDs[i]), 12);
			mString = strcat(mString, "...");
		}	
		else
			mString = strcpy(mString, GetAppName(appIDs[i]));
		sprintf(working, ": %lu", OO_GetFlashAppSize(appIDs[i]));
		mString = strcat(mString, working);
		PopupAddText(menu, -1, mString, 0);
		memset(mString, '\0', strlen(mString));
	}
	selected = PopupDo(menu, CENTER, CENTER, 0);
	HeapFree(menu);
	if (selected)
	{
		selected -= 1;
		sprintf(mString, "App ID: %d\nApp Size: %lu bytes\nMarked for Delete: %d", appIDs[selected], OO_GetFlashAppSize(appIDs[selected]), (int)OO_appIsMarkedDelete(appIDs[selected]));
		dlgKey = DlgMessage(GetAppName(appIDs[selected]), mString, BT_OK, BT_CANCEL);
		free(mString);
		free(appIDs);
		if (dlgKey != KEY_ENTER)
			doFlashApps();
	}
	else
	{
		free(mString);
		free(appIDs);		
	}
}

//Main Function
void _main(void)
{
	int argtype;
	short selected = 0;
	static HANDLE menu = H_NULL;
	ESI argptr;
	
	//If there is command line input, get and integer and assign it to selected
	if (ArgCount() == 1)
	{
		InitArgPtr(argptr);
		while ((argtype = GetArgType(argptr)) != END_TAG)
		{
			if (argtype == POSINT_TAG)
				selected = GetIntArg(argptr);
			else
				ER_throw(ER_ARGUMENT);
		}
	}
	//If there is not, create a menu for input
	else
	{
		menu = PopupNew("Memory Usage", 0);
		PopupAddText(menu, -1, "Show file/folder mem", 0);
		PopupAddText(menu, -1, "Show flash app mem", 0);
		PopupAddText(menu, -1, "Show general info", 0);
		PopupAddText(menu, -1, "Garbage collect", 0);
		selected = PopupDo(menu, CENTER, CENTER, 0);
		HeapFree(menu);
	}

	//Launch a function depending on the input
	if (selected == 1)
		doFilesAndFolders();
	else if (selected == 2)
		doFlashApps();
	else if (selected == 3)
		doGeneralInformation();
	else if (selected == 4)
		EM_GC(TRUE);
}
